package ssw.mj.test;

import java.io.StringReader;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

import junit.framework.TestCase;
import ssw.mj.Errors;
import ssw.mj.Messages;
import ssw.mj.Parser;
import ssw.mj.Scanner;
import ssw.mj.Token;
import ssw.mj.codegen.Code;
import ssw.mj.codegen.Decoder;
import ssw.mj.symtab.Tab;

/**
 * Base class for test cases with utility methods used by all tests.
 */
public class CompilerTestCase extends TestCase {
	public static final String CR = "\r";
	public static final String LF = "\n";

	/** stores the number of the current exercise for simple access by the test cases */
	public static final int EX = Configuration.CURRENT_EXERCISE;
	
	private List<String> expectedErrors;
	private List<String> expectedTokens;
	private List<String> expectedSymTab;
	private byte[] expectedCode;
	private int expectedCodeSize;
	private int expectedDataSize;
	private int expectedMainPC;
	
	
	@Override
	protected void setUp() {
		// reset the static fields of all classes to their default values
		resetClass(Errors.class);
		resetClass(Scanner.class);
		resetClass(Parser.class);
		resetClass(Tab.class);
		resetClass(Code.class);
		
		// initialize expected compiler output
		expectedErrors = new ArrayList<String>();
		expectedTokens = new ArrayList<String>();
		expectedSymTab = new ArrayList<String>();
		expectedCode = new byte[8192];
		expectedCodeSize = 0;
		expectedDataSize = 0;
		expectedMainPC = -1;

		// print header for console output
		System.out.println("--------------------------------------------------");
		System.out.println(getName());
	}

	/**
	 * reset all static fields of the given class to their default values (null or 0)
	 * @param name Name of the class to reset
	 */
	private void resetClass(Class klass) {
//		Class klass = null;
//		try {
//			klass = Class.forName(name);
//		} catch (ClassNotFoundException ex) {
//			throw new Error("class " + name + " not found");
//		}
		
		try {
			Field[] fields = klass.getDeclaredFields();
			for (Field field : fields) {
				if (Modifier.isStatic(field.getModifiers()) && !Modifier.isFinal(field.getModifiers())) {
					field.setAccessible(true);
					Class fieldKlass = field.getType();
					if (fieldKlass == int.class) {
						field.setInt(null, 0);
					} else if (fieldKlass == short.class) {
						field.setShort(null, (short)0);
					} else if (fieldKlass == byte.class) {
						field.setByte(null, (byte)0);
					} else if (fieldKlass == char.class) {
						field.setChar(null, (char)0);
					} else if (fieldKlass == long.class) {
						field.setLong(null, 0);
					} else if (fieldKlass == float.class) {
						field.setFloat(null, 0);
					} else if (fieldKlass == double.class) {
						field.setDouble(null, 0);
					} else if (fieldKlass == boolean.class) {
						field.setBoolean(null, false);
					} else {
						field.set(null, null);
					}
				}
			}
		} catch (Exception ex) {
			ex.printStackTrace();
			throw new Error("Could not reset class " + klass.getName(), ex);
		}
	}
	

	protected void init(String s) {
		// use indirection via config to break dependency on class Scanner
		Scanner.init(new StringReader(s));
	}


	private List<String> splitString(String s) {
		StringTokenizer st = new StringTokenizer(s, "\n");
		List<String> result = new ArrayList<String>();
		while (st.hasMoreTokens()) {
			result.add(st.nextToken());
		}
		return result;
	}
	
	private void printAndAssert(String title, List<String> expected, List<String> actual) {
		if (Configuration.PRINT_ALL_OUTPUT || !expected.equals(actual)) {
			System.out.println(title);
			System.out.format("  %-60s %s\n", "expected", "actual");
			int lines = Math.max(expected.size(), actual.size());
			for (int i = 0; i < lines; i++) {
				String expectedLine = (i < expected.size() ? expectedLine = expected.get(i) : "");
				String actualLine =   (i < actual.size()   ? actualLine =   actual.get(i)   : "");
				System.out.format("%s %-60s %s\n", (expectedLine.equals(actualLine) ? " " : "x"), expectedLine, actualLine);
			}
		}
		
		assertEquals(title, expected, actual);
	}

	protected void scanAndVerify() {
		List<String> actualTokens = new ArrayList<String>();
		// scan only the expected number of tokens to prevent endless loops
		for (int i = 0; i < getExpectedTokens().size(); i++) {
			// use indirection via config to break dependency on class Scanner
			actualTokens.add(Scanner.next().toString());
		}

		printAndAssert("Errors", getExpectedErrors(), getActualErrors());
		printAndAssert("Tokens", getExpectedTokens(), actualTokens);
		assertTrue("Complete Input Scanned", Scanner.next().toString().contains("end of file"));
	}

	protected void parseAndVerify() {
		// use indirection via config to break dependency on class Parser
		boolean panic = false;
		try {
			Parser.parse();
		} catch (Error error) {
			if (EX <= 3 && Messages.getString("PANIC_MODE").equals(error.getMessage())) {
				panic = true;
			} else {
				throw error;
			}
		}
		
		if (EX <= 3) {
			assertEquals("panic mode error", getExpectedErrors().size() > 0, panic);
		}

		// print SymTab and Code in Java syntax for easy test case generation
//		System.out.println();
//		for (String s : getActualSymTab()) {
//			System.out.println("\t\texpectSymTab(\"" + s + "\");");
//		}
//		System.out.println();
//		List<String> actual = getActualCode();
//		for (int i = 0; i < actual.size() - 3; i++) {
//			String s = actual.get(i);
//			System.out.println("\t\texpectCode(EX >= 6, Code." + s.substring(s.indexOf(":") + 2, s.length()) + ");");
//		}
//		System.out.println("\t\texpectDataSize(" + Code.dataSize + ");");
//		System.out.println("\t\texpectMainPC(" + Code.mainpc + ");");
//		System.out.println();
		
		printAndAssert("Errors", getExpectedErrors(), getActualErrors());
		if (expectedSymTab.size() > 0) {
			printAndAssert("Symbol Table", getExpectedSymTab(), getActualSymTab());
		}
		if (expectedCodeSize > 0) {
			printAndAssert("Code Buffer", getExpectedCode(), getActualCode());
		}
		
		if (EX >= 4) {
			// with panic mode, the complete input is not scanned, so cannot check this
			assertTrue("Complete Input Scanned", Scanner.next().toString().contains("end of file"));
		}
	}


	private List<String> getExpectedErrors() {
		return expectedErrors;
	}
	
	private List<String> getActualErrors() {
		return splitString(Errors.dump());
	}
	
	private List<String> getExpectedTokens() {
		return expectedTokens;
	}
	
	private List<String> getExpectedSymTab() {
		return expectedSymTab;
	}
	
	private List<String> getActualSymTab() {
		return splitString(Tab.dump());
	}
	
	private List<String> getExpectedCode() {
		List<String> result = splitString(Decoder.decode(expectedCode, 0, expectedCodeSize));
		result.add("#CodeSize: " + expectedCodeSize);
		result.add("#DataSize: " + expectedDataSize);
		result.add("#MainPC: " + expectedMainPC);
		return result;
	}

	private List<String> getActualCode() {
		return splitString(Code.dump());
	}
	
	
	protected void expectError(boolean enabled, int line, int col, String msgKey, Object... msgParams) {
		if (enabled) {
			expectedErrors.add("-- line " + line + " col " + col + ": " + Messages.getString(msgKey, msgParams));
		}
	}

	
	protected void expectToken(int kind, int line, int col) {
		expectedTokens.add("line " + line + ", col " + col + ", kind " + kind + " (" + Token.kindToString(kind) + ")");
	}
	protected void expectToken(int kind, int line, int col, String str) {
		expectedTokens.add("line " + line + ", col " + col + ", kind " + kind + " (" + Token.kindToString(kind) + ")" + ", str " + str);
	}
	protected void expectToken(int kind, int line, int col, int val) {
		expectedTokens.add("line " + line + ", col " + col + ", kind " + kind + " (" + Token.kindToString(kind) + ")" + ", val " + val);
	}
	protected void expectToken(int kind, int line, int col, char val) {
		expectedTokens.add("line " + line + ", col " + col + ", kind " + kind + " (" + Token.kindToString(kind) + ")" + ", val '" + val + "'");
	}
	
	
	protected void expectSymTab(String line) {
		if (EX >= 4) {
			expectedSymTab.add(line);
		}
	}

	protected void expectSymTabUniverse() {
		// first part of the symbol table (universe) that is equal for all programs
		expectSymTab("-- begin scope (0 variables) --");
		expectSymTab("Type int: int");
		expectSymTab("Type char: char");
		expectSymTab("Constant: class(0) null = 0");
		expectSymTab("Method: char chr(1) at 0");
		expectSymTab("  Local Variable 0: int i");
		expectSymTab("Method: int ord(1) at 0");
		expectSymTab("  Local Variable 0: char ch");
		expectSymTab("Method: int len(1) at 0");
		expectSymTab("  Local Variable 0: void[] arr");
	}


	private void expectCode1(int x) {
		expectedCode[expectedCodeSize++] = (byte) x;
	}
	private void expectCode2(int x) {
		expectCode1(x >> 8); 
		expectCode1(x); 
	}

	private void expectCode4(int x) {
		expectCode2(x >> 16); 
		expectCode2(x); 
	}
	
	protected void expectCode(boolean enabled, int op, Object... params) {
		if (enabled) {
			expectCode1(op);
			for (Object param : params) {
				if (param instanceof Byte) {
					expectCode1((Byte)param);
				} else if (param instanceof Short) {
					expectCode2((Short)param);
				} else if (param instanceof Integer) {
					expectCode4((Integer)param);
				} else {
					throw new Error("invalid parameter");
				}
			}
		}
	}

	protected void expectDataSize(int size) {
		if (EX >= 6) {
			expectedDataSize = size;
		}
	}

	protected void expectMainPC(int pc) {
		if (EX >= 6) {
			expectedMainPC = pc;
		}
	}
}
